#include <WinSock2.h>
#include <Windows.h>
#include <boost/noncopyable.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include "double_click_monitor.h"
#include "url_processor.h"

class double_click_monitor_impl
{
public:
	double_click_monitor_impl(uint32_t default_loop_timeout)
		:click_timer_(click_service_, boost::posix_time::seconds(default_loop_timeout)),
		skip_next_timer_(false)
	{
	}

	boost::asio::io_service click_service_;
	boost::asio::deadline_timer click_timer_;
	bool skip_next_timer_;

	std::wstring last_clicked_url_;
};

double_click_monitor::double_click_monitor()
	:impl_(new double_click_monitor_impl(default_loop_timeout))
{
}

double_click_monitor::~double_click_monitor()
{
	delete impl_;
}

void double_click_monitor::run()
{
	impl_->click_timer_.async_wait(boost::bind(&double_click_monitor::on_timer, this, url_info(), boost::asio::placeholders::error));
	impl_->click_service_.run();
}

void double_click_monitor::on_timer(const url_info& info, const boost::system::error_code& e)
{
	if(!info.url.empty() && !impl_->skip_next_timer_ && impl_->last_clicked_url_ == info.url)
	{
		if(!e)
		{
			url_processor::set_work_item(info.url, info.x, info.y, true);
		}
		else if(e == boost::asio::error::operation_aborted)
		{
			ShellExecuteW(0, L"open", info.url.c_str(), 0, 0, SW_SHOW);
			impl_->skip_next_timer_ = true;
		}
	}

	if(e != boost::asio::error::operation_aborted)
	{
		impl_->skip_next_timer_ = false;
		impl_->click_timer_.expires_from_now(boost::posix_time::seconds(default_loop_timeout));
		impl_->click_timer_.async_wait(boost::bind(&double_click_monitor::on_timer, this, url_info(), boost::asio::placeholders::error));
	}
}

void double_click_monitor::on_reset_timer(uint32_t duration, const url_info& info)
{
	impl_->last_clicked_url_ = info.url;
	impl_->click_timer_.cancel();
	impl_->click_timer_.expires_from_now(boost::posix_time::milliseconds(duration));
	impl_->click_timer_.async_wait(boost::bind(&double_click_monitor::on_timer, this, info, boost::asio::placeholders::error));
}

void double_click_monitor::reset_timer(uint32_t duration, const std::wstring& url, int32_t x, int32_t y)
{
	impl_->click_service_.post(boost::bind(&double_click_monitor::on_reset_timer, this, duration, url_info(url, x, y)));
}

double_click_monitor::url_info::url_info()
	:x(0), y(0)
{
}

double_click_monitor::url_info::url_info(const std::wstring& page_url, int32_t mousex, int32_t mousey)
	:x(mousex), y(mousey), url(page_url)
{
}
