#include <Windows.h>
#include <string>
#include <intrin.h>
#include <CommCtrl.h>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/date_time.hpp>
#include "web_content_loader.h"
#include "com_initializer.h"
#include "gdi_plus_initializer.h"
#include "image_window.h"
#include "url_processor.h"
#include "minhook_initializer.h"
#include "configuration.h"
#include "utility.h"
#include "dialogs.h"
#include "skype.h"
#include "menu.h"
#include "resource.h"
#include "double_click_monitor.h"
#include "../MinHook/libMinHook/MinHook.h"

typedef BOOL (WINAPI* DrawMenuBar_func)(HWND hWnd);

typedef BOOL (WINAPI *ExtTextOutW_func)(HDC hdc, int X, int Y,
	UINT fuOptions, const RECT *lprc, LPCWSTR lpString, UINT cbCount, const INT *lpDx);

typedef BOOL (WINAPI* PeekMessageW_func)(LPMSG lpMsg, HWND hWnd, UINT wMsgFilterMin,
	UINT wMsgFilterMax, UINT wRemoveMsg);

typedef HINSTANCE (WINAPI* ShellExecuteW_func)(HWND hwnd, LPCWSTR lpOperation, LPCWSTR lpFile, LPCWSTR lpParameters,
	LPCWSTR lpDirectory, INT nShowCmd);

namespace
{
ExtTextOutW_func ExtTextOutW_original = 0;
DrawMenuBar_func DrawMenuBar_original = 0;
PeekMessageW_func PeekMessageW_original = 0;
ShellExecuteW_func ShellExecuteW_original = 0;

boost::mutex draw_menu_bar_mutex;

HINSTANCE shell_execute_success = reinterpret_cast<HINSTANCE>(50);

//Skype hangs if boost::asio objects are created as static,
//so create them later
std::auto_ptr<double_click_monitor> monitor;

std::auto_ptr<image_window> displayer;
}

void click_monitor_thread()
{
	monitor->run();
}

LRESULT CALLBACK menu_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR /*uIdSubclass*/, DWORD_PTR /*dwRefData*/)
{
	switch(uMsg)
	{
	case WM_COMMAND:
		if(HIWORD(wParam) != 0)
			break;

		if(LOWORD(wParam) == skype::settings_menu_id)
			dialogs::show_settings_dialog(hWnd);
		else if(LOWORD(wParam) == skype::about_menu_id)
			dialogs::show_about_dialog(hWnd);

		break;

	case WM_NCDESTROY:
		RemoveWindowSubclass(hWnd, menu_proc, skype::subclass_id);
		break;
	}

	return DefSubclassProc(hWnd, uMsg, wParam, lParam);
}

HINSTANCE WINAPI DetourShellExecuteW(HWND hwnd, LPCWSTR lpOperation, LPCWSTR lpFile, LPCWSTR lpParameters,
	LPCWSTR lpDirectory, INT nShowCmd)
{
	if(utility::get_module_name_from_address(_ReturnAddress()) == skype::module_name
		&& configuration::get_property("plugin_on", true)
		&& configuration::get_property("preview_on_click", false)
		&& lpOperation == std::wstring(L"open")
		&& lpFile
		&& utility::is_http_https_url(lpFile))
	{
		utility::mouse_info mouse(utility::get_mouse_info());
		monitor->reset_timer(GetDoubleClickTime(), lpFile, mouse.x, mouse.y);
		return shell_execute_success;
	}

	return ShellExecuteW_original(hwnd, lpOperation, lpFile, lpParameters, lpDirectory, nShowCmd);
}

BOOL WINAPI DetourPeekMessageW(LPMSG lpMsg, HWND hWnd, UINT wMsgFilterMin, UINT wMsgFilterMax, UINT wRemoveMsg)
{
	HWND win = utility::get_main_window(GetCurrentProcessId());
	if(win && GetCurrentThreadId() == GetWindowThreadProcessId(win, 0))
	{
		DrawMenuBar(win);

		MH_DisableHook(&PeekMessageW);
	}

	return PeekMessageW_original(lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax, wRemoveMsg);
}

BOOL WINAPI DetourDrawMenuBar(HWND hWnd)
{
	try
	{
		std::wstring module(utility::get_module_name_from_address(_ReturnAddress()));
		if(module == skype::module_name
			|| module == utility::get_module_name_from_address(configuration::get_instance()))
		{
			if(hWnd && GetCurrentThreadId() == GetWindowThreadProcessId(hWnd, 0))
			{
				DWORD_PTR refdata = 0;
				if(!GetWindowSubclass(hWnd, menu_proc, skype::subclass_id, &refdata))
					SetWindowSubclass(hWnd, menu_proc, skype::subclass_id, 0);
			}

			{
				boost::mutex::scoped_lock lock(draw_menu_bar_mutex);

				if(!menu::has_menu(hWnd, skype::main_menu_text))
				{
					menu::submenu_list submenus;
					submenus.push_back(menu::submenu(skype::settings_menu_id, skype::settings_menu_text));
					submenus.push_back(menu::submenu(skype::about_menu_id, skype::about_menu_text));
					menu::add_menu(hWnd, skype::main_menu_text, submenus);
				}
			}
		}
	}
	catch(const std::exception&)
	{
	}

	return DrawMenuBar_original(hWnd);
}

BOOL WINAPI DetourExtTextOutW(HDC hdc, int X, int Y,
	UINT fuOptions, const RECT *lprc, LPCWSTR lpString, UINT cbCount, const INT *lpDx)
{
	try
	{
		if(utility::get_module_name_from_address(_ReturnAddress()) == skype::module_name
			&& configuration::get_property("plugin_on", true)
			&& !configuration::get_property("preview_on_click", false))
		{
			utility::mouse_info mouse(utility::get_mouse_info());

			if(mouse.is_shown
				&& !mouse.left_pressed && !mouse.middle_pressed && !mouse.right_pressed)
			{
				std::wstring item(lpString, cbCount);

				if(utility::is_http_https_url(item))
				{
					if(*(lpString + cbCount) != 0)
						item = lpString; //Read till nullbyte

					url_processor::set_work_item(item, mouse.x, mouse.y, mouse.is_hand);
				}
			}
		}
	}
	catch(const std::exception&)
	{
	}

	return ExtTextOutW_original(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);
}

void CALLBACK active_window_changed(HWINEVENTHOOK hWinEventHook,
	DWORD event,
	HWND hwnd,
	LONG idObject,
	LONG idChild,
	DWORD dwEventThread,
	DWORD dwmsEventTime)
{
	if(displayer.get())
		displayer->hide();
}

void main_thread()
{
#ifndef EXE_TEST
	monitor.reset(new double_click_monitor());
	boost::thread click_thread(click_monitor_thread);

	minhook_initializer minhook;

	if(MH_CreateHook(&ShellExecuteW, &DetourShellExecuteW, reinterpret_cast<void**>(&ShellExecuteW_original)) != MH_OK)
		return;

	if(MH_EnableHook(&ShellExecuteW) != MH_OK)
		return;

	if(MH_CreateHook(&ExtTextOutW, &DetourExtTextOutW, reinterpret_cast<void**>(&ExtTextOutW_original)) != MH_OK)
		return;

	if(MH_EnableHook(&ExtTextOutW) != MH_OK)
		return;

	if(MH_CreateHook(&DrawMenuBar, &DetourDrawMenuBar, reinterpret_cast<void**>(&DrawMenuBar_original)) != MH_OK)
		return;

	if(MH_EnableHook(&DrawMenuBar) != MH_OK)
		return;

	if(MH_CreateHook(&PeekMessageW, &DetourPeekMessageW, reinterpret_cast<void**>(&PeekMessageW_original)) != MH_OK)
		return;

	if(MH_EnableHook(&PeekMessageW) != MH_OK)
		return;
#endif

	com_initializer com;
	gdi_plus_initializer gdi;
	utility::init_common_controls();

	web_content_loader loader;
	displayer.reset(new image_window(new unpacked_image(new Gdiplus::Bitmap(1, 1))));

	configuration::set_loader_and_displayer(loader, *displayer);
	configuration::apply_settings();

	displayer->start();
	HWINEVENTHOOK hook = SetWinEventHook(EVENT_SYSTEM_FOREGROUND, EVENT_SYSTEM_FOREGROUND,
		0, active_window_changed, 0, 0, WINEVENT_OUTOFCONTEXT);

	url_processor::set_wait_image(
		utility::load_from_resource(MAKEINTRESOURCEW(IDR_WAIT_GIF), RT_RCDATA, false).get());

	url_processor::start(loader, *displayer);

	displayer->message_queue();
	url_processor::wait();

#ifndef EXE_TEST
	if(click_thread.joinable())
		click_thread.join();

	if(hook)
		UnhookWinEvent(hook);

	MH_DisableHook(&ExtTextOutW);
	MH_DisableHook(&DrawMenuBar);
	MH_DisableHook(&ShellExecuteW);
	MH_DisableHook(&PeekMessageW);
#endif
}

void check_properties()
{
	if(!configuration::get_property("connection_timeout", skype::default_connection_timeout))
		configuration::replace_property("connection_timeout", skype::default_connection_timeout, true);
	
	if(!configuration::get_property("read_timeout", skype::default_read_timeout))
		configuration::replace_property("read_timeout", skype::default_read_timeout, true);
	
	if(!configuration::get_property("max_data_size", skype::default_max_data_size))
		configuration::replace_property("max_data_size", skype::default_max_data_size, true);
	
	if(!configuration::get_property("max_image_x_y_size", skype::default_max_image_x_y_size))
		configuration::replace_property("max_image_x_y_size", skype::default_max_image_x_y_size, true);
	
	if(!configuration::get_property("max_gif_x_y_size", skype::default_max_gif_x_y_size))
		configuration::replace_property("max_gif_x_y_size", skype::default_max_gif_x_y_size, true);
	
	if(!configuration::get_property("max_title_length", skype::default_max_title_length))
		configuration::replace_property("max_title_length", skype::default_max_title_length, true);

	{
		utility::encoding_list encodings(utility::get_supported_encoding_list());
		if(std::find(encodings.begin(), encodings.end(),
			configuration::get_property("default_encoding", skype::default_encoding)) == encodings.end())
		{
			configuration::replace_property("default_encoding", skype::default_encoding, true);
		}
	}
}

#ifndef EXE_TEST
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID)
{
	switch(fdwReason)
	{
	case DLL_PROCESS_ATTACH:
		configuration::set_instance(hinstDLL);

		try
		{
			configuration::load(skype::settings_file_name, skype::plugin_name);
			check_properties();
		}
		catch(const std::exception&)
		{
		}

		boost::thread worker(main_thread);
		worker.detach();
		break;
	}

	return TRUE;
}
#else
int CALLBACK WinMain(HINSTANCE hInstance, HINSTANCE /*hPrevInstance*/, LPSTR /*lpCmdLine*/, int /*nCmdShow*/)
{
	configuration::set_instance(hInstance);

	try
	{
		configuration::load(skype::settings_file_name, skype::plugin_name);
		check_properties();
	}
	catch(const std::exception&)
	{
	}

#if 1
	utility::init_common_controls();
	dialogs::show_settings_dialog(0);
	//return 0;
#endif

	//TODO:      
	//TODO: cache last url?

	boost::thread([] () {
		Sleep(2000);
		url_processor::set_work_item(L"http://img0.joyreactor.cc/pics/post/%D0%B3%D0%B8%D1%84%D0%BA%D0%B0-%D0%BE%D0%BD%D0%B0%D0%BD%D0%B8%D0%B7%D0%BC-%D0%B1%D1%83%D0%B4%D1%83%D1%89%D0%B5%D0%B5-1138466.gif", 200, 200, true);
		Sleep(5000);
	});

	main_thread();

	return 0;
}
#endif